# Copyright 2019-2020 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_machine import AbstractMachine
from netket.graph import AbstractGraph
import numpy as _np

from numba import (
    jit,
    float64,
    complex128,
)


@jit(fastmath=True)
def _log_cosh_sum(x, out, add_factor=None):
    x = x * _np.sign(x.real)
    if add_factor is None:
        for i in range(x.shape[0]):
            out[i] = _np.sum(x[i] - _np.log(2.0) + _np.log(1.0 + _np.exp(-2.0 * x[i])))
    else:
        for i in range(x.shape[0]):
            out[i] += add_factor * (
                _np.sum(x[i] - _np.log(2.0) + _np.log(1.0 + _np.exp(-2.0 * x[i])))
            )
    return out


class RbmSpin(AbstractMachine):
    r"""
    A fully connected Restricted Boltzmann Machine (RBM). This type of
    RBM has spin 1/2 hidden units and is defined by:

    .. math:: \Psi(s_1,\dots s_N) = e^{\sum_i^N a_i s_i} \times \Pi_{j=1}^M \cosh
     \left(\sum_i^N W_{ij} s_i + b_j \right)

    for arbitrary local quantum numbers :math:`s_i`. Here a and b are
    called, respectively, visible and hidden bias.

    The weights can be taken to be complex-valued (default option) or real-valued.
    """

    def __init__(
        self,
        hilbert,
        n_hidden=None,
        alpha=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        automorphisms=None,
        dtype=complex,
    ):
        r"""
        Constructs a new ``RbmSpin`` machine:

        Args:
            hilbert: Hilbert space object for the system.
            n_hidden: The number of hidden spin units. If n_hidden=None, the number
                   of hidden units is deduced from the parameter alpha.
            alpha: `alpha * hilbert.size` is the number of hidden spins.
                   If alpha=None, the number of hidden units must
                   be explicitely set in the argument n_hidden.
            use_visible_bias: If ``True`` a bias on the visible units is taken.
                              Default ``True``.
            use_hidden_bias: If ``True`` bias on the hidden units is taken.
                             Default ``True``.
            automorphisms (optional): a list of automorphisms to use as symmetries.
            dtype: either complex or float, is the type used for the weights.

        Examples:
            A ``RbmSpin`` machine with hidden unit density
            alpha = 2 for a one-dimensional L=20 spin-half system:

            >>> from netket.machine import RbmSpin
            >>> from netket.hilbert import Spin
            >>> hi = Spin(s=0.5, total_sz=0, N=20)
            >>> ma = RbmSpin(hilbert=hi,alpha=2)
            >>> print(ma.n_par)
            860
        """

        super().__init__(hilbert, dtype=dtype)

        n = hilbert.size

        if dtype is not float and dtype is not complex:
            raise TypeError("dtype must be either float or complex")

        self._npdtype = _np.complex128 if dtype is complex else _np.float64

        self._autom, self.n_hidden, alpha_symm = self._get_hidden(
            automorphisms, hilbert, n_hidden, alpha
        )

        m = self.n_hidden

        self._w = _np.empty((n, m), dtype=self._npdtype)
        self._a = _np.empty(n, dtype=self._npdtype) if use_visible_bias else None
        self._b = _np.empty(m, dtype=self._npdtype) if use_hidden_bias else None
        self._r = _np.empty((1, m), dtype=self._npdtype)

        self._n_bare_par = (
            self._w.size
            + (self._a.size if self._a is not None else 0)
            + (self._b.size if self._b is not None else 0)
        )

        if automorphisms is None or automorphisms is False:
            self._ws = self._w.view()
            self._as = self._a.view() if use_visible_bias else None
            self._bs = self._b.view() if use_hidden_bias else None
            self._n_par = self._n_bare_par
        else:
            self._der_mat_symm, self._n_par = self._build_der_mat(
                use_visible_bias, use_hidden_bias, n, alpha_symm, self._autom
            )

            self._ws = _np.empty((n, alpha_symm), dtype=self._npdtype)
            self._as = _np.empty(1, dtype=self._npdtype) if use_visible_bias else None
            self._bs = (
                _np.empty(alpha_symm, dtype=self._npdtype) if use_hidden_bias else None
            )

            self._set_bare_parameters(
                self._a, self._b, self._w, self._as, self._bs, self._ws, self._autom
            )

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self._n_par

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The
                 length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        x = x.astype(dtype=self._npdtype)

        return self._log_val_kernel(x, out, self._w, self._a, self._b, self._r)

    @staticmethod
    @jit(nopython=True)
    def _log_val_kernel(x, out, W, a, b, r):

        if x.ndim != 2:
            raise RuntimeError("Invalid input shape, expected a 2d array")

        if out is None:
            out = _np.empty(x.shape[0], dtype=_np.complex128)
        r = x.dot(W)
        if b is None:
            _log_cosh_sum(r, out)
        else:
            _log_cosh_sum(r + b, out)

        if a is not None:
            out = out + x.dot(a)

        return out

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        if self._autom is None:
            return self._bare_der_log(x, out)
        else:
            self._outb = self._bare_der_log(x)
            if out is None:
                out = _np.empty((x.shape[0], self._n_par), dtype=_np.complex128)
            return _np.matmul(self._outb, self._der_mat_symm, out=out)

    def _bare_der_log(self, x, out=None):

        if x.ndim != 2:
            raise RuntimeError("Invalid input shape, expected a 2d array")

        if out is None:
            out = _np.empty((x.shape[0], self._n_bare_par), dtype=_np.complex128)

        batch_size = x.shape[0]
        n_visible = x.shape[1]

        i = 0
        if self._a is not None:
            out[:, i : i + n_visible] = x
            i += n_visible

        r = self._r
        r = _np.dot(x, self._w)
        if self._b is not None:
            r += self._b
        r = _np.tanh(r)

        if self._b is not None:
            out[:, i : i + self.n_hidden] = r
            i += self.n_hidden

        t = out[:, i : i + self._w.size]
        t.shape = (batch_size, self._w.shape[0], self._w.shape[1])
        _np.einsum("ij,il->ijl", x, r, out=t)

        return out

    @property
    def state_dict(self):
        return self.state_dict_with_prefix()

    def state_dict_with_prefix(self, prefix=""):
        r"""A dictionary containing the parameters of this machine"""
        from collections import OrderedDict

        od = OrderedDict()
        if self._dtype is complex:
            if self._a is not None:
                od[prefix + "a"] = self._as.view()

            if self._b is not None:
                od[prefix + "b"] = self._bs.view()

            od[prefix + "w"] = self._ws.view()
        else:
            if self._a is not None:
                self._ac = self._as.astype(_np.complex128)
                self._as = self._ac.real.view()
                od[prefix + "a"] = self._ac.view()
                if self._autom is None:
                    self._a = self._as.view()

            if self._b is not None:
                self._bc = self._bs.astype(_np.complex128)
                self._bs = self._bc.real.view()
                od[prefix + "b"] = self._bc.view()
                if self._autom is None:
                    self._b = self._bs.view()

            self._wc = self._ws.astype(_np.complex128)
            self._ws = self._wc.real.view()
            od[prefix + "w"] = self._wc.view()
            if self._autom is None:
                self._w = self._ws.view()

        return od

    @property
    def parameters(self):
        return _np.concatenate(tuple(p.reshape(-1) for p in self.state_dict.values()))

    @parameters.setter
    def parameters(self, p):
        if p.shape != (self.n_par,):
            raise ValueError(
                "p has wrong shape: {}; expected ({},)".format(p.shape, self.n_par)
            )

        i = 0
        for x in map(lambda x: x.reshape(-1), self.state_dict.values()):
            _np.copyto(x, p[i : i + x.size])
            i += x.size

        if self._autom is not None:
            self._set_bare_parameters(
                self._a, self._b, self._w, self._as, self._bs, self._ws, self._autom
            )

    @staticmethod
    @jit
    def _build_der_mat(use_visible_bias, use_hidden_bias, n_visible, alpha, permtable):
        perm_size = permtable.shape[0]
        n_hidden = alpha * perm_size

        n_par = int(n_visible * alpha + use_visible_bias + alpha * use_hidden_bias)
        n_bare_par = int(
            n_visible * n_hidden
            + use_visible_bias * n_visible
            + n_hidden * use_hidden_bias
        )

        der_mat_symm = _np.zeros((n_par, n_bare_par))

        k = 0
        k_bare = 0

        if use_visible_bias:
            # derivative with respect to a
            for p in range(n_visible):
                der_mat_symm[k, p] = 1
                k_bare += 1
            k += 1

        if use_hidden_bias:
            # derivatives with respect to b
            for p in range(n_hidden):
                k_symm = int(_np.floor(p / perm_size))
                der_mat_symm[k_symm + k, k_bare] = 1
                k_bare += 1

            k += alpha

        # derivatives with respect to W
        for i in range(n_visible):
            for j in range(n_hidden):
                isymm = permtable[j % perm_size, i]
                jsymm = int(_np.floor(j / perm_size))
                ksymm = jsymm + int(alpha * isymm)

                der_mat_symm[ksymm + k, k_bare] = 1

                k_bare += 1

        return der_mat_symm.T, n_par

    @staticmethod
    def _get_hidden(automorphisms, hilbert, n_hidden, alpha):
        if (automorphisms is None) or (automorphisms is False):
            if alpha is None:
                m = n_hidden
                if n_hidden < 0:
                    raise RuntimeError("n_hidden must be positive.")
            else:
                m = int(round(alpha * hilbert.size))
                if alpha < 0:
                    raise ValueError("`alpha` should be non-negative")

            if n_hidden is not None:
                if n_hidden != m:
                    raise RuntimeError(
                        """n_hidden is inconsistent with the given alpha.
                                       Remove one of the two or provide consistent values."""
                    )

            return None, m, m
        else:
            if isinstance(automorphisms, AbstractGraph):
                autom = _np.asarray(automorphisms.automorphisms())
            else:
                try:
                    autom = _np.asarray(automorphisms)
                    assert hilbert.size == autom.shape[1]
                except:
                    raise RuntimeError("Cannot find a valid automorphism array.")

            if not _np.issubdtype(type(alpha), _np.integer):
                raise ValueError(
                    "alpha must be a positive integer value when using symmetries."
                )

            alphasym = int(alpha * autom.shape[1] / autom.shape[0])
            if alphasym == 0 and alpha > 0:
                print(
                    "Warning, the given value of alpha is too small, given the size of the automorphisms."
                )
            m = int(alphasym * autom.shape[0])

            return autom, m, alphasym

    @staticmethod
    @jit
    def _set_bare_parameters(a, b, W, a_s, b_s, W_s, permtable):

        perm_size = permtable.shape[0]

        if a is not None:
            a.fill(a_s)

        if b is not None:
            for j in range(b.shape[0]):
                jsymm = int(_np.floor(j / perm_size))
                b[j] = b_s[jsymm]

        for i in range(W.shape[0]):
            for j in range(W.shape[1]):
                jsymm = int(_np.floor(j / perm_size))
                W[i, j] = W_s[permtable[j % perm_size][i], jsymm]


class RbmSpinReal(RbmSpin):
    r"""
    A fully connected Restricted Boltzmann Machine (RBM) with real-valued parameters.
    See RbmSpin for more details.

    """

    def __init__(
        self,
        hilbert,
        n_hidden=None,
        alpha=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        automorphisms=None,
    ):
        r"""
        Constructs a new ``RbmSpinReal`` machine:

        Args:
            hilbert: Hilbert space object for the system.
            n_hidden: Number of hidden units.
            alpha: Hidden unit density.
            use_visible_bias: If ``True`` then there would be a
                            bias on the visible units.
                            Default ``True``.
            use_hidden_bias: If ``True`` then there would be a
                           bias on the visible units.
                           Default ``True``.
            automorphisms (optional): A list of automorphisms to be taken as symmetries of the
                           rbm. They can be obtained by calling graph.automorphism()

        Examples:
           A ``RbmSpinReal`` machine with hidden unit density
           alpha = 2 for a one-dimensional L=20 spin-half system:


           >>> from netket.machine import RbmSpinReal
           >>> from netket.hilbert import Spin
           >>> g = Hypercube(length=20, n_dim=1)
           >>> hi = Spin(s=0.5, total_sz=0, N=4)
           >>> ma = RbmSpinReal(hilbert=hi,alpha=2)
           >>> print(ma.n_par)
           860
        """
        super().__init__(
            hilbert,
            n_hidden=n_hidden,
            alpha=alpha,
            use_visible_bias=use_visible_bias,
            use_hidden_bias=use_hidden_bias,
            automorphisms=automorphisms,
            dtype=float,
        )


class RbmSpinSymm(RbmSpin):
    r"""
    A fully connected Restricted Boltzmann Machine with lattice
    symmetries. This type of RBM has spin 1/2 hidden units and is
    defined by:

    .. math:: \Psi(s_1,\dots s_N) = e^{\sum_i^N a_i s_i} \times \Pi_{j=1}^M
    \cosh \left(\sum_i^N W_{ij} s_i + b_j \right)

    for arbitrary local quantum numbers :math:`s_i`. However, the weights
    (:math:`W_{ij}`) and biases (:math:`a_i`, :math:`b_i`) respects the
    symmetries of the lattice as specificed in hilbert.graph.automorphisms.

    The values of the weights can be chosen to be complex or real-valued.

    """

    def __init__(
        self,
        hilbert,
        automorphisms,
        alpha=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        dtype=complex,
    ):
        r"""
        Constructs a new ``RbmSpinSymm`` machine with complex weights:

        Args:
           hilbert: Hilbert space object for the system.
           alpha: Hidden unit density.
           use_visible_bias: If ``True`` then there would be a
                            bias on the visible units.
                            Default ``True``.
           use_hidden_bias: If ``True`` then there would be a
                           bias on the visible units.
                           Default ``True``.

        Examples:
           A ``RbmSpinSymm`` machine with hidden unit density
           alpha = 2 for a one-dimensional L=20 spin-half system:

           >>> from netket.machine import RbmSpinSymm
           >>> from netket.hilbert import Spin
           >>> hi = Spin(s=0.5, total_sz=0, N=20)
           >>> ma = RbmSpinSymm(hilbert=hi, alpha=2)
           >>> print(ma.n_par)
           43
        """
        super().__init__(
            hilbert,
            alpha=alpha,
            use_visible_bias=use_visible_bias,
            use_hidden_bias=use_hidden_bias,
            automorphisms=automorphisms,
            dtype=dtype,
        )


class RbmMultiVal(RbmSpin):
    r"""
    A fully connected Restricted Boltzmann Machine suitable for large local hilbert spaces.
    Local quantum numbers are passed through a one hot encoding that maps them onto
    an enlarged space of +/- 1 spins. In turn, these quantum numbers are used with a
    standard RbmSpin wave function.
    """

    def __init__(
        self,
        hilbert,
        n_hidden=None,
        alpha=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        automorphisms=None,
        dtype=complex,
    ):

        r"""
        Constructs a new ``RbmMultiVal`` machine:

        Args:
            hilbert: Hilbert space object for the system.
            n_hidden: The number of hidden spin units. If n_hidden=None, the number
                   of hidden units is deduced from the parameter alpha.
            alpha: `alpha * hilbert.size` is the number of hidden spins.
                   If alpha=None, the number of hidden units must
                   be explicitely set in the argument n_hidden.
            use_visible_bias: If ``True`` then there would be a
                            bias on the visible units.
                            Default ``True``.
            use_hidden_bias: If ``True`` then there would be a
                           bias on the visible units.
                           Default ``True``.
            automorphisms (optional): A list of automorphisms to be taken as symmetries of the
                           rbm. They can be obtained by calling graph.automorphism()
            dtype: either complex or float, is the type used for the weights.

        Examples:
            A ``RbmMultiVal`` machine with hidden unit density
            alpha = 1 for a one-dimensional bosonic system:

            >>> from netket.machine import RbmMultiVal
            >>> from netket.hilbert import Boson
            >>> hi = Boson(n_max=3, n_bosons=8, N=10)
            >>> ma = RbmMultiVal(hilbert=hi, alpha=1, dtype=float, use_visible_bias=False)
            >>> print(ma.n_par)
            1056
        """
        local_states = _np.asarray(hilbert.local_states, dtype=int)

        # creating a fictious hilbert object to pass to the standard rbm
        l_hilbert = _np.zeros(local_states.size * hilbert.size)

        # creating the symmetries for the extended space
        autom = self._make_extended_symmetry(automorphisms, hilbert)

        super().__init__(
            l_hilbert,
            n_hidden,
            alpha,
            use_visible_bias,
            use_hidden_bias,
            autom,
            dtype,
        )

        self.hilbert = hilbert
        self._local_states = local_states

    @staticmethod
    @jit
    def _one_hot(vec, local_states):
        if vec.ndim != 2:
            raise RuntimeError("Invalid input shape, expected a 2d array")

        # one hotting and converting to -/+ 1
        res = (
            vec.reshape((vec.shape[1], vec.shape[0], -1)) == local_states
        ) * 2.0 - 1.0
        return res.reshape((vec.shape[0], -1))

    @staticmethod
    def _make_extended_symmetry(automorphisms, hilbert):
        if automorphisms is not None:
            if isinstance(automorphisms, AbstractGraph):
                autom = _np.asarray(automorphisms.automorphisms())
            else:
                try:
                    autom = _np.asarray(automorphisms)
                    assert hilbert.size == autom.shape[1]
                except:
                    raise RuntimeError("Cannot find a valid automorphism array.")

            ldim = len(hilbert.local_states)

            autop = _np.zeros((autom.shape[0], autom.shape[1] * ldim), dtype=int)

            for k in range(autom.shape[0]):
                for i in range(autom.shape[1]):
                    autop[k, i * ldim : (i + 1) * ldim] = _np.arange(
                        autom[k, i] * ldim, (autom[k, i] + 1) * ldim, 1, dtype=int
                    )
        else:
            autop = None
        return autop

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The
                 length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        return super().log_val(self._one_hot(x, self._local_states), out)

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        return super().der_log(self._one_hot(x, self._local_states), out)


class RbmSpinPhase(AbstractMachine):
    r"""
    A fully connected Restricted Boltzmann Machine (RBM) with real-valued parameters.
    In this case, two RBMs are taken to parameterize, respectively, phase
    and amplitude of the wave-function, as introduced in Torlai et al., Nature Physics 14, 447–450(2018).
    This type of RBM has spin 1/2 hidden units and is defined by:

    .. math:: \Psi(s_1,\dots s_N) = e^{\sum_i^N a_i s_i} \times \Pi_{j=1}^M
            \cosh \left(\sum_i^N W_{ij} s_i + b_j \right)

    for arbitrary local quantum numbers :math:`s_i`.
    """

    def __init__(
        self,
        hilbert,
        alpha=None,
        n_hidden=None,
        alpha_a=None,
        alpha_p=None,
        n_hidden_a=None,
        n_hidden_p=None,
        use_visible_bias=True,
        use_hidden_bias=True,
        automorphisms=None,
    ):
        r"""
        Constructs a new ``RbmSpinPhase`` machine:

        Args:
            hilbert: Hilbert space object for the system.
            alpha: Hidden unit density.
            n_hidden: The number of hidden spin units to be used in both RBMs. If None,
                     n_hidden_a and n_hidden_p must be specified.
            alpha_a: Hidden unit density for the amplitude. If None, the general parameter alpha is used.
            alpha_p: Hidden unit density for the phase. If None, the general parameter alpha is used.
                    n_hidden_a: The number of hidden spin units to be used to represent the amplitude.
            If None, the general parameter alpha (or n_hidden) is used to derive the number of hidden units.
            n_hidden_p: The number of hidden spin units to be used to represent the phase.
            If None, the general parameter alpha (or n_hidden) is used to derive the number of hidden units.
            use_visible_bias: If ``True`` then there would be a
                            bias on the visible units.
                            Default ``True``.
            use_hidden_bias: If ``True`` then there would be a
                           bias on the visible units.
                           Default ``True``.
            automorphisms (optional): List of automorphisms to enforcee symmetries in the Boltzmann Machine.

        Examples:

        """
        if n_hidden is not None:
            n_hidden_a = n_hidden
            n_hidden_p = n_hidden

        if alpha is not None:
            alpha_a = alpha
            alpha_p = alpha

        self._rbm_a = RbmSpin(
            hilbert=hilbert,
            n_hidden=n_hidden_a,
            alpha=alpha_a,
            use_visible_bias=use_visible_bias,
            use_hidden_bias=use_hidden_bias,
            automorphisms=automorphisms,
            dtype=float,
        )
        self._rbm_p = RbmSpin(
            hilbert=hilbert,
            n_hidden=n_hidden_p,
            alpha=alpha_p,
            use_visible_bias=use_visible_bias,
            use_hidden_bias=use_hidden_bias,
            automorphisms=automorphisms,
            dtype=float,
        )

        self._n_par = self._rbm_a.n_par + self._rbm_p.n_par

        super().__init__(hilbert, dtype=float)

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self._n_par

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The
                 length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        out = self._rbm_a.log_val(x, out)
        out += 1.0j * self._rbm_p.log_val(x)

        return out

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        n_par = self._n_par
        n_par_a = self._rbm_a.n_par

        if out is None:
            out = _np.empty((x.shape[0], n_par), dtype=_np.complex128)

        self._rbm_a.der_log(x, out[:, :n_par_a])

        self._rbm_p.der_log(x, out[:, n_par_a:])
        out[:, n_par_a:] *= 1.0j

        return out

    @property
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        od = self._rbm_a.state_dict_with_prefix(prefix="amplitude_")
        od.update(self._rbm_p.state_dict_with_prefix(prefix="phase_"))
        return od

    @property
    def parameters(self):
        return _np.concatenate(tuple(p.reshape(-1) for p in self.state_dict.values()))

    @parameters.setter
    def parameters(self, p):
        self._rbm_a.parameters = p[: self._rbm_a.n_par]
        self._rbm_p.parameters = p[self._rbm_a.n_par :]



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from functools import partial
from netket.stats import subtract_mean
from netket.utils import mpi
from netket.jax import tree_conj, tree_axpy

# Stochastic Reconfiguration with jvp and vjp

# Here O (Oks) is the jacobian (derivatives w.r.t. params) of the vectorised (in x) log wavefunction (forward_fn) evaluated at all samples.
# instead of computing (and storing) the full jacobian matrix jvp and vjp are used to implement the matrix vector multiplications with it.
# Expectation values are then just the mean over the leading dimension.


def O_jvp(forward_fn, params, samples, v):
    # TODO apply the transpose of mpi_sum_jax(x)[0] (allreduce) to v here
    # in order to get correct transposition with MPI
    _, res = jax.jvp(lambda p: forward_fn(p, samples), (params,), (v,))
    return res


def O_vjp(forward_fn, params, samples, w):
    _, vjp_fun = jax.vjp(forward_fn, params, samples)
    res, _ = vjp_fun(w)
    return jax.tree_map(lambda x: mpi.mpi_sum_jax(x)[0], res)  # allreduce w/ MPI.SUM


def OH_w(forward_fn, params, samples, w):
    r"""
    compute  Oᴴw
    (where ᴴ denotes the hermitian transpose)
    """

    # Oᴴw = (wᴴO)ᴴ
    # The transposition of the 1D arrays is omitted in the implementation:
    # (wᴴO)ᴴ -> (w* O)*

    return tree_conj(O_vjp(forward_fn, params, samples, w.conjugate()))


def Odagger_O_v(forward_fn, params, samples, v, *, center=False):
    r"""
    if center=False (default):
        compute ⟨O†O⟩v

    else (center=True):
        compute ⟨O†ΔO⟩v
        where ΔO = O-⟨O⟩
    """

    # w is an array of size n_samples; each MPI rank has its own slice
    w = O_jvp(forward_fn, params, samples, v)
    # w /= n_samples (elementwise):
    w = w * (1.0 / (samples.shape[0] * mpi.n_nodes))

    if center:
        w = subtract_mean(w)  # w/ MPI

    return OH_w(forward_fn, params, samples, w)


Odagger_DeltaO_v = partial(Odagger_O_v, center=True)


# TODO block the computations (in the same way as done with MPI) if memory consumtion becomes an issue
def mat_vec(v, forward_fn, params, samples, diag_shift):
    r"""
    compute (S + diag_shift) v

    where the elements of S are given by Sₖₗ = ⟨Oₖ†ΔOₗ⟩
    where ΔOₖ = Oₖ-⟨Oₖ⟩
    and Oₖ (operator) is derivative of the log wavefunction w.r.t parameter k
    The expectation values are calculated as mean over the samples

    v: a pytree with the same structure as params
    forward_fn(params, x): a vectorised function returning the logarithm of the wavefunction for each configuration in x
    params: pytree of parameters with arrays as leaves
    samples: an array of samples (when using MPI each rank has its own slice of samples)
    diag_shift: a scalar diagonal shift
    """

    res = Odagger_DeltaO_v(forward_fn, params, samples, v)
    # add diagonal shift:
    res = tree_axpy(diag_shift, v, res)  # res += diag_shift * v
    return res



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class AdaMax(AbstractOptimizer):
    r"""AdaMax Optimizer.

    AdaMax is an adaptive stochastic gradient descent method,
    and a variant of `Adam <https://arxiv.org/pdf/1412.6980.pdf>`_ based on the infinity norm.
    In contrast to the SGD, AdaMax offers the important advantage of being much
    less sensitive to the choice of the hyper-parameters (for example, the learning rate).

    Given a stochastic estimate of the gradient of the cost function :math:`G(\mathbf{p})`,
    AdaMax performs an update:

    .. math:: p^{\prime}_k = p_k + \mathcal{S}_k,

    where :math:`\mathcal{S}_k` implicitly depends on all the history of the optimization up to the current point.
    The NetKet naming convention of the parameters strictly follows the one introduced by the authors of AdaMax.
    For an in-depth description of this method, please refer to
    `Kingma, D., & Ba, J. (2015). Adam: a method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_
    (Algorithm 2 therein).
    """

    def __init__(self, alpha=0.001, beta1=0.9, beta2=0.999, epscut=1.0e-7):
        r"""
        Constructs a new ``AdaMax`` optimizer.

        Args:
            alpha: The step size.
            beta1: First exponential decay rate.
            beta2: Second exponential decay rate.
            epscut: Small epsilon cutoff.

        Examples:
            Simple AdaMax optimizer.

            >>> from netket.optimizer import AdaMax
            >>> op = AdaMax()
        """

        if epscut <= 0:
            raise ValueError("Invalid epsilon cutoff.")
        if alpha < 0:
            raise ValueError("Invalid alpha.")
        if beta1 < 0 or beta1 > 1:
            raise ValueError("Invalid beta1.")
        if beta2 < 0 or beta2 > 1:
            raise ValueError("Invalid beta1.")

        self._alpha = alpha
        self._beta1 = beta1
        self._beta2 = beta2
        self._epscut = epscut

        self._mt = None
        self._ut = None
        self._niter = 0

    def update(self, grad, pars):
        if self._mt is None:
            self._mt = _np.zeros(pars.shape[0])
            self._ut = _np.zeros(pars.shape[0])

        self._mt = self._beta1 * self._mt + (1.0 - self._beta1) * grad

        self._ut = _np.maximum(
            _np.maximum(_np.abs(grad), self._beta2 * self._ut), self._epscut
        )

        self._niter += 1

        eta = self._alpha / (1.0 - self._beta1 ** self._niter)

        pars -= eta * self._mt / self._ut

        return pars

    def reset(self):
        if self._mt is not None:
            self._mt.fill(0.0)
            self._ut.fill(0.0)
            self._niter = 0

    def __repr__(self):
        rep = "AdaMax optimizer with these parameters :"
        rep += "\nStep size alpha = " + str(self._alpha)
        rep += "\nbeta1 = " + str(self._beta1)
        rep += "\nbeta2 = " + str(self._beta2)
        rep += "\nepscut = " + str(self._epscut)
        return rep



import pytest
import netket.legacy as nk

if not nk.utils.jax_available:
    pytest.skip("skipping jax-only SR-onthefly tests", allow_module_level=True)

import jax
import jax.numpy as jnp
import jax.flatten_util
import numpy as np
from jax.scipy.sparse.linalg import cg
from netket.legacy.optimizer.jax._sr_onthefly import *
from netket.legacy.optimizer.jax.stochastic_reconfiguration import (
    _jax_cg_solve_onthefly,
)

pytestmark = pytest.mark.legacy


@partial(jax.vmap, in_axes=(None, 0))
def f(params, x):
    return (
        params["a"][0] * x[0]
        + params["b"] * x[1]
        + params["c"] * (x[0] * x[1])
        + jnp.sin(x[1] * params["a"][1])
    )


n_samp = 25
samples = jnp.array(np.random.random((n_samp, 2)))

params = {
    "a": jnp.array([1.0 + 1.2j, -4.0 + -0.7j]),
    "b": jnp.array(2.0),
    "c": jnp.array(-0.55 + 4.33j),
}

v = {
    "a": jnp.array([0.7 + 1.1j, -3.9 + 3.5j]),
    "b": jnp.array(0.3),
    "c": jnp.array(-0.74 + 3j),
}

grad = {
    "a": jnp.array([0.01 + 23.0j, 1.99 + -7.2j]),
    "b": jnp.array(-0.231),
    "c": jnp.array(1.22 - 0.45j),
}

v_tilde = jnp.array(np.random.random(n_samp) + 1j * np.random.random(n_samp))


def flatten(x):
    x_flat, _ = jax.flatten_util.ravel_pytree(x)
    return x_flat


def toreal(x):
    if jnp.iscomplexobj(x):
        return jnp.array([x.real, x.imag])
    else:
        return x


def tree_toreal(x):
    return jax.tree_map(toreal, x)


def tree_toreal_flat(x):
    return flatten(tree_toreal(x))


# invert the transformation tree_toreal_flat using linear_transpose (AD)
def reassemble_complex(x, fun=tree_toreal_flat, target=params):
    # target: a tree with the expected shape and types of the result
    (res,) = jax.linear_transpose(fun, target)(x)
    res = tree_conj(res)
    # fix the dtypes:
    return tree_cast(res, target)


def f_real_flat(p, samples):
    return f(reassemble_complex(p), samples)


def f_real_flat_scalar(params, x):
    return f_real_flat(params, jnp.expand_dims(x, 0))[0]


# same as in nk.machine.Jax R->C
@partial(jax.vmap, in_axes=(None, 0))
def grads_real(params, x):
    r = jax.grad(lambda pars, v: f_real_flat_scalar(pars, v).real)(params, x)
    i = jax.grad(lambda pars, v: f_real_flat_scalar(pars, v).imag)(params, x)
    return jax.lax.complex(r, i)


params_real_flat = tree_toreal_flat(params)
grad_real_flat = tree_toreal_flat(grad)
v_real_flat = tree_toreal_flat(v)

ok_real = grads_real(params_real_flat, samples)
okmean_real = ok_real.mean(axis=0)
dok_real = ok_real - okmean_real
S_real = (dok_real.conjugate().transpose() @ dok_real / n_samp).real


def tree_allclose(t1, t2):
    t = jax.tree_multimap(jnp.allclose, t1, t2)
    return all(jax.tree_util.tree_flatten(t)[0])


def test_reassemble_complex():
    assert tree_allclose(params, reassemble_complex(tree_toreal_flat(params)))


def test_vjp():
    actual = O_vjp(samples, params, v_tilde, f)
    expected = tree_conj(reassemble_complex((v_tilde @ ok_real).real))
    assert tree_allclose(actual, expected)


def test_mean():
    actual = O_mean(samples, params, f)
    expected = tree_conj(reassemble_complex(okmean_real.real))
    assert tree_allclose(actual, expected)


def test_OH_w():
    actual = OH_w(samples, params, v_tilde, f)
    expected = reassemble_complex((ok_real.conjugate().transpose() @ v_tilde).real)
    assert tree_allclose(actual, expected)


def test_jvp():
    actual = O_jvp(samples, params, v, f)
    expected = ok_real @ v_real_flat
    assert tree_allclose(actual, expected)


def test_Odagger_DeltaO_v():
    actual = Odagger_DeltaO_v(samples, params, v, f)
    expected = reassemble_complex(S_real @ v_real_flat)
    assert tree_allclose(actual, expected)


def test_matvec():
    diag_shift = 0.01
    actual = mat_vec(v, f, params, samples, diag_shift)
    expected = reassemble_complex(S_real @ v_real_flat + diag_shift * v_real_flat)
    assert tree_allclose(actual, expected)


def test_matvec_linear_transpose():
    w = v
    (actual,) = jax.linear_transpose(
        lambda v_: mat_vec(v_, f, params, samples, 0.0), v
    )(w)
    # use that S is hermitian:
    # S^T = (O^H O)^T = O^T O* = (O^H O)* = S*
    # S^T w = S* w = (S w*)*
    expected = tree_conj(mat_vec(tree_conj(w), f, params, samples, 0.0))
    # (expected,) = jax.linear_transpose(lambda v_: reassemble_complex(S_real @ tree_toreal_flat(v_)), v)(v)
    assert tree_allclose(actual, expected)


def test_cg():
    # also tests if matvec can be jitted and be differentiated with AD
    diag_shift = 0.001
    sparse_tol = 1.0e-5
    sparse_maxiter = None
    actual = _jax_cg_solve_onthefly(
        v, f, params, samples, grad, diag_shift, sparse_tol, sparse_maxiter
    )

    def mv_real(v):
        return S_real @ v + diag_shift * v

    expected = reassemble_complex(
        cg(
            mv_real,
            grad_real_flat,
            x0=v_real_flat,
            tol=sparse_tol,
            maxiter=sparse_maxiter,
        )[0]
    )
    assert tree_allclose(actual, expected)



from functools import partial
from netket.legacy.stats import sum_inplace as _sum_inplace

import jax
import jax.numpy as jnp

from jax import jit
from jax.scipy.sparse.linalg import cg
from jax.tree_util import tree_flatten
from netket.legacy.vmc_common import jax_shape_for_update
from netket.utils.mpi import n_nodes, mpi4jax_available

from ._sr_onthefly import mat_vec as _mat_vec_onthefly, tree_cast as _tree_cast


def _S_grad_mul(oks, v):
    r"""
    Computes y = 1/N * ( O^\dagger * O * v ) where v is a vector of
    length n_parameters, and O is a matrix (n_samples, n_parameters)
    """
    # TODO apply the transpose of sum_inplace (allreduce) to v here
    # in order to get correct transposition with MPI
    v_tilde = jnp.matmul(oks, v) / (oks.shape[0] * n_nodes)
    y = jnp.matmul(oks.conjugate().transpose(), v_tilde)
    return _sum_inplace(y)


def _compose_result_cmplx(v, y, diag_shift):
    return v * diag_shift + y


def _compose_result_real(v, y, diag_shift):
    return (v * diag_shift + y).real


def _matvec_cmplx(v, oks, diag_shift):
    y = _S_grad_mul(oks, v)
    return _compose_result_cmplx(v, y, diag_shift)


def _matvec_real(v, oks, diag_shift):
    y = _S_grad_mul(oks, v)
    return _compose_result_real(v, y, diag_shift)


@partial(jit, static_argnums=1)
def _jax_cg_solve(x0, mat_vec, oks, grad, diag_shift, sparse_tol, sparse_maxiter):
    r"""
    Solves the SR flow equation using the conjugate gradient method
    """

    _mat_vec = partial(mat_vec, oks=oks, diag_shift=diag_shift)

    out, _ = cg(_mat_vec, grad, x0=x0, tol=sparse_tol, maxiter=sparse_maxiter)

    return out


@partial(jit, static_argnums=1)
def _jax_cg_solve_onthefly(
    x0,
    forward_fn,
    params,
    samples,
    grad,
    diag_shift,
    sparse_tol,
    sparse_maxiter,
):
    _mat_vec = partial(
        _mat_vec_onthefly,
        forward_fn=forward_fn,
        params=params,
        samples=samples,
        diag_shift=diag_shift,
    )
    out, _ = cg(_mat_vec, grad, x0=x0, tol=sparse_tol, maxiter=sparse_maxiter)
    return out


@jit
def _flatten_grad_and_oks(grads, jac):
    r"""Reshapes grads and jac from tree like structures to arrays if jax_available

    Args:
        grads,jac: pytrees of jax arrays or numpy array

    Returns:
        A 1D array of gradients and a 2D array of the jacobian
    """

    grads = jnp.concatenate(tuple(fd.reshape(-1) for fd in tree_flatten(grads)[0]))
    jac = jnp.concatenate(
        tuple(fd.reshape(len(fd), -1) for fd in tree_flatten(jac)[0]), -1
    )
    return grads, jac


@jit
def _subtract_mean_from_oks(oks):
    return oks - _sum_inplace(jnp.sum(oks, axis=0) / (oks.shape[0] * n_nodes))


class SR:
    r"""
    Performs stochastic reconfiguration (SR) updates.
    """

    def __init__(
        self,
        machine,
        lsq_solver=None,
        diag_shift=0.01,
        use_iterative=True,
        svd_threshold=None,
        sparse_tol=None,
        sparse_maxiter=None,
        onthefly=True,
    ):

        if n_nodes > 1 and not mpi4jax_available:
            raise RuntimeError(
                """
                Cannot use Jax-Stochastic Reconfiguration with multiple MPI processes unless mpi4jax package is installed.
                Please run `pip install mpi4jax` and restart python.
                """
            )

        self._onthefly = onthefly
        self._lsq_solver = lsq_solver
        self._diag_shift = diag_shift
        self._use_iterative = use_iterative
        self._has_complex_parameters = None
        self._machine = None

        # Quantities for sparse solver
        self.sparse_tol = 1.0e-5 if sparse_tol is None else sparse_tol
        self.sparse_maxiter = sparse_maxiter
        self._lsq_solver = lsq_solver
        self._x0 = None
        self._mat_vec = None
        self._init_solver()

        if not self._use_iterative:
            raise NotImplementedError("JaxSR supports only iterative solvers")

        if machine is not None:
            self.setup(machine)

    def _init_solver(self):
        lsq_solver = self._lsq_solver

        if lsq_solver in ["gmres", "cg", "minres", "jaxcg"]:
            self._use_iterative = True
        if lsq_solver in ["ColPivHouseholder", "QR", "SVD", "Cholesky"]:
            self._use_iterative = False

        if self._use_iterative:
            if lsq_solver is None or lsq_solver == "cg":
                self._lsq_solver = "cg"
            elif lsq_solver in ["gmres", "minres"]:
                raise Warning(
                    "Conjugate gradient is the only sparse solver currently implemented in Jax. Defaulting to cg"
                )
                self._lsq_solver = "cg"
            else:
                raise RuntimeError("Unknown sparse lsq_solver " + lsq_solver + ".")

    def setup(self, machine):
        r"""
        Sets up this Sr object to work with the selected machine.
        This mainly sets internal flags `has_complex_parameters` and the
        method used to flatten/unflatten the gradients.

        Args:
            machine: the machine
        """
        self._machine = machine
        self._has_complex_parameters = machine.has_complex_parameters

        if self._has_complex_parameters:
            self._mat_vec = _matvec_cmplx
        else:
            self._mat_vec = _matvec_real

    def compute_update(self, oks, grad, out=None):
        r"""
        Solves the SR flow equation for the parameter update ẋ.

        The SR update is computed by solving the linear equation
           Sẋ = f
        where S is the covariance matrix of the partial derivatives
        O_i(v_j) = ∂/∂x_i log Ψ(v_j) and f is a generalized force (the loss
        gradient).

        Args:
            oks: A pytree of the jacobians ∂/∂x_i log Ψ(v_j)
            grad: A pytree of the forces f
            out: A pytree of the parameter updates that will be ignored
        """

        if self.has_complex_parameters is None or self._machine is None:
            raise ValueError("This SR object is not properly initialized.")

        # if using MPI the number of samples on each rank has to be the same

        grad, oks = _flatten_grad_and_oks(grad, oks)
        oks = _subtract_mean_from_oks(oks)

        n_par = grad.shape[0]

        if self._x0 is None:
            if self.has_complex_parameters:
                self._x0 = jnp.zeros(n_par, dtype=jnp.complex128)
            else:
                self._x0 = jnp.zeros(n_par, dtype=jnp.float64)

        if self.has_complex_parameters:
            if self._use_iterative:
                if self._lsq_solver == "cg":
                    out = _jax_cg_solve(
                        self._x0,
                        self._mat_vec,
                        oks,
                        grad,
                        self._diag_shift,
                        self.sparse_tol,
                        self.sparse_maxiter,
                    )
                self._x0 = out
        else:
            if self._use_iterative:
                if self._lsq_solver == "cg":
                    out = _jax_cg_solve(
                        self._x0,
                        self._mat_vec,
                        oks,
                        grad.real,
                        self._diag_shift,
                        self.sparse_tol,
                        self.sparse_maxiter,
                    )
                self._x0 = out

        out = jax_shape_for_update(out, self._machine.parameters)

        return out

    def compute_update_onthefly(self, samples, grad, out=None):
        r"""
        Solves the SR flow equation for the parameter update ẋ.

        The SR update is computed by solving the linear equation
           Sẋ = f
        where S is the covariance matrix of the partial derivatives
        O_i(v_j) = ∂/∂x_i log Ψ(v_j) and f is a generalized force (the loss
        gradient).

        Args:
            samples: An array of samples
            grad: A pytree of the forces f
            out: A pytree of the parameter updates that will be ignored
        """

        # cast the leaves of grad to the same dtypes as the parameters
        grad = _tree_cast(grad, self._machine.parameters)

        if self._x0 is None:
            self._x0 = jax.tree_map(jnp.zeros_like, grad)  # x0 = jnp.zeros_like(grad)

        if self._use_iterative:
            if self._lsq_solver == "cg":
                out = _jax_cg_solve_onthefly(
                    self._x0,
                    self._machine._forward_fn_nj,
                    self._machine.parameters,
                    samples,
                    grad,
                    self._diag_shift,
                    self.sparse_tol,
                    self.sparse_maxiter,
                )
            self._x0 = out

        return out

    def __repr__(self):
        rep = "SR(solver="

        if self._use_iterative:
            rep += "iterative"

        rep += ", has_complex_parameters=" + str(self._has_complex_parameters) + ")"
        return rep

    def info(self, depth=0):
        indent = " " * 4 * depth
        rep = indent
        rep += "Stochastic reconfiguration method for "
        rep += (
            "complex-parameters" if self._has_complex_parameters else "real-parameters"
        )
        rep += " wavefunctions\n"

        rep += indent + "Solver: "

        if self._use_iterative:
            rep += "iterative (Conjugate Gradient)"

        rep += "\n"

        return rep

    @property
    def has_complex_parameters(self):
        return self._has_complex_parameters

    @property
    def onthefly(self):
        return self._onthefly



import pytest
import netket as nk


@pytest.fixture
def _mpi_size(request):
    """
    Fixture returning the number of MPI nodes detected by NetKet
    """
    return nk.utils.mpi.n_nodes


@pytest.fixture
def _mpi_rank(request):
    """
    Fixture returning the MPI rank detected by NetKet
    """
    return nk.utils.mpi.rank


@pytest.fixture
def _mpi_comm(request):
    """
    Fixture returning the MPI communicator used by NetKet
    """

    from netket.utils.mpi import MPI_py_comm

    return MPI_py_comm


def pytest_addoption(parser):
    parser.addoption(
        "--legacy",
        action="store_true",
        dest="legacy",
        default=False,
        help="enable legacy tests",
    )

    parser.addoption(
        "--legacy-only",
        action="store_true",
        dest="only_legacy",
        default=False,
        help="enable legacy tests and disable everything else",
    )


def pytest_configure(config):
    if not config.option.legacy:
        setattr(config.option, "markexpr", "not legacy")

    if config.option.only_legacy:
        setattr(config.option, "markexpr", "legacy")

    if nk.utils.mpi.n_nodes > 1:
        setattr(config.option, "markexpr", "not legacy")



from functools import singledispatch
from . import numpy


@singledispatch
def SR(
    machine,
    lsq_solver=None,
    diag_shift=0.01,
    use_iterative=True,
    svd_threshold=None,
    sparse_tol=None,
    sparse_maxiter=None,
):
    return numpy.SR(
        machine,
        lsq_solver,
        diag_shift,
        use_iterative,
        svd_threshold,
        sparse_tol,
        sparse_maxiter,
    )



from numba import jit
import numpy as _np
from netket.legacy import random as _random


class _CustomKernel:
    def __init__(self, move_operators, move_weights=None):

        self._rand_op_n = _np.empty(1, dtype=_np.intp)
        self._sections = _np.empty(1, dtype=_np.intp)
        self._x_prime = _np.empty(1)
        self._mels = _np.empty(1)
        self._get_conn = move_operators.get_conn_filtered
        self._n_operators = move_operators.n_operators

        if move_weights is None:
            self._move_weights = _np.ones(self._n_operators, dtype=_np.float64)
        else:
            self._move_weights = _np.asarray(move_weights, dtype=_np.float64)

        self._check_operators(move_operators.operators)

        # Check move weights
        if self._move_weights.shape != (self._n_operators,):
            raise ValueError("move_weights have the wrong shape")
        if self._move_weights.min() < 0:
            raise ValueError("move_weights must be positive")

        # normalize the probabilities and compute the cumulative
        self._move_weights /= self._move_weights.sum()
        self._move_cumulative = _np.cumsum(self._move_weights)

        self._hilbert = move_operators.hilbert

    def _check_operators(self, operators):
        for op in operators:
            assert op.imag.max() < 1.0e-10
            assert op.min() >= 0
            assert _np.allclose(op.sum(axis=0), 1.0)
            assert _np.allclose(op.sum(axis=1), 1.0)
            assert _np.allclose(op, op.T)

    def transition(self, state, state_1, log_prob_corr):

        self._rand_op_n, self._sections = self._pick_random_and_init(
            state.shape[0], self._move_cumulative, self._rand_op_n, self._sections
        )

        self._x_prime, self._mels = self._get_conn(
            state, self._sections, self._rand_op_n
        )

        self._choose_and_return(
            state_1, self._x_prime, self._mels, self._sections, log_prob_corr
        )

    @staticmethod
    @jit(nopython=True)
    def _pick_random_and_init(batch_size, move_cumulative, out, sections):

        if out.size != batch_size:
            out = _np.empty(batch_size, dtype=out.dtype)
            sections = _np.empty(batch_size, dtype=out.dtype)

        for i in range(batch_size):
            p = _random.uniform()
            out[i] = _np.searchsorted(move_cumulative, p)
        return out, sections

    @staticmethod
    @jit(nopython=True)
    def _choose_and_return(state_1, x_prime, mels, sections, log_prob_corr):
        low = 0
        for i in range(state_1.shape[0]):
            p = _random.uniform()
            exit_state = 0
            cumulative_prob = mels[low].real
            while p > cumulative_prob:
                exit_state += 1
                cumulative_prob += mels[low + exit_state].real
            state_1[i] = x_prime[low + exit_state]
            low = sections[i]

        log_prob_corr.fill(0.0)

    def random_state(self, state):

        for i in range(state.shape[0]):
            self._hilbert.random_state(out=state[i])



import numpy as _np
from .abstract_sampler import AbstractSampler
from netket.legacy.machine.density_matrix import AbstractDensityMatrix
from netket.hilbert import DoubledHilbert
import netket.legacy.random


class ExactSampler(AbstractSampler):
    r"""
    This sampler generates i.i.d. samples from $$|\Psi(s)|^2$$.
    In order to perform exact sampling, $$|\Psi(s)|^2$$ is precomputed an all
    the possible values of the quantum numbers $$s$$. This sampler has thus an
    exponential cost with the number of degrees of freedom, and cannot be used
    for large systems, where Metropolis-based sampling are instead a viable
    option.
    """

    def __init__(self, machine, sample_size=16):
        r"""
        Constructs a new ``ExactSampler`` given a machine.

        Args:
            machine: A machine $$\Psi(s)$$ used for the sampling.
                     The probability distribution being sampled
                     from is $$F(\Psi(s))$$, where the function
                     $$F(X)$$, is arbitrary, by default $$F(X)=|X|^2$$.

            sample_size: The number of independent samples to be generated at each invocation of __next__.
        """
        super().__init__(machine, sample_size)
        if isinstance(machine, AbstractDensityMatrix):
            self.hilbert = DoubledHilbert(machine.hilbert)
        else:
            self.hilbert = machine.hilbert
        self._machine_pow = 2.0
        self.reset()

    def reset(self, init_random=False):
        self._prob = _np.absolute(self.machine.to_array()) ** self.machine_pow
        self._prob /= self._prob.sum()

    def __next__(self):
        numbers = netket.legacy.random.choice(
            self._prob.size, size=self.sample_shape[0], replace=True, p=self._prob
        )
        return self.hilbert.numbers_to_states(numbers)

    def generate_samples(self, n_samples, init_random=False, samples=None):

        if samples is None:
            samples = _np.zeros((n_samples, self.sample_shape[0], self.sample_shape[1]))

        numbers = netket.legacy.random.choice(
            self._prob.size,
            size=self.sample_shape[0] * n_samples,
            replace=True,
            p=self._prob,
        )
        samples[:] = self.hilbert.numbers_to_states(numbers).reshape(samples.shape)

        return samples

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self._machine_pow = m_power
        self.reset()



from functools import partial
import numpy as _np
from scipy.linalg import lstsq as _lstsq
from scipy.linalg import cho_factor as _cho_factor
from scipy.linalg import cho_solve as _cho_solve
from scipy.sparse.linalg import LinearOperator
from netket.legacy.stats import sum_inplace as _sum_inplace, mean as _mean
from scipy.sparse.linalg import cg, gmres, minres
from netket.utils.mpi import (
    MPI_py_comm as _MPI_comm,
    n_nodes as _n_nodes,
)


class SR:
    r"""
    Performs stochastic reconfiguration (SR) updates.
    """

    def __init__(
        self,
        machine,
        lsq_solver=None,
        diag_shift=0.01,
        use_iterative=True,
        svd_threshold=None,
        sparse_tol=None,
        sparse_maxiter=None,
    ):

        self._lsq_solver = lsq_solver
        self._diag_shift = diag_shift
        self._use_iterative = use_iterative
        self._has_complex_parameters = None
        self._svd_threshold = svd_threshold
        self._scale_invariant_pc = False
        self._S = None
        self._last_rank = None

        # Temporary arrays
        self._v_tilde = None
        self._res_t = None

        # Quantities for sparse solver
        self.sparse_tol = 1.0e-5 if sparse_tol is None else sparse_tol
        self.sparse_maxiter = sparse_maxiter
        self._lsq_solver = lsq_solver
        self._x0 = None
        self._init_solver()

        self._comm = _MPI_comm

        if machine is not None:
            self.setup(machine)

    def _init_solver(self):
        lsq_solver = self._lsq_solver

        if lsq_solver in ["gmres", "cg", "minres"]:
            self._use_iterative = True
        if lsq_solver in ["ColPivHouseholder", "QR", "SVD", "Cholesky"]:
            self._use_iterative = False

        if self._use_iterative:
            if lsq_solver is None:
                lsq_solver = "gmres" if self._has_complex_parameters else "minres"

            if lsq_solver == "gmres":
                self._sparse_solver = partial(gmres, atol="legacy")
            elif lsq_solver == "cg":
                self._sparse_solver = partial(cg, atol="legacy")
            elif lsq_solver == "minres":
                if self._has_complex_parameters:
                    raise RuntimeError(
                        "minres can be used only for real-valued parameters."
                    )
                self._sparse_solver = minres
            else:
                raise RuntimeError("Unknown sparse lsq_solver " + lsq_solver + ".")

        else:
            if (
                lsq_solver is None
                or "ColPivHouseholder" in lsq_solver
                or "QR" in lsq_solver
            ):
                self._lapack_driver = "gelsy"
            elif "SVD" in lsq_solver or "Cholesky" in lsq_solver:
                self._lapack_driver = None
            else:
                self._lapack_driver = None
                raise RuntimeError("Unknown lsq_solver" + lsq_solver + ".")

        if self._use_iterative and self._svd_threshold is not None:
            raise ValueError(
                "The svd_threshold option is available only for non-sparse solvers."
            )

    def setup(self, machine):
        """
        Sets up this Sr object to work with the selected machine.
        This mainly sets internal flags `has_complex_parameters` and the
        method used to flatten/unflatten the gradients.

        Args:
            machine: the machine

        """
        self._has_complex_parameters = machine.has_complex_parameters
        self._init_solver()

    def compute_update(self, oks, grad, out=None):
        r"""
        Solves the SR flow equation for the parameter update ẋ.

        The SR update is computed by solving the linear equation
           Sẋ = f
        where S is the covariance matrix of the partial derivatives
        O_i(v_j) = ∂/∂x_i log Ψ(v_j) and f is a generalized force (the loss
        gradient).

        Args:
            oks: The matrix of log-derivatives,
                O_i(v_j)
            grad: The vector of forces f.
            out: Output array for the update ẋ.
        """

        oks -= _mean(oks, axis=0)

        if self.has_complex_parameters is None:
            raise ValueError(
                "has_complex_parameters not set: this SR object is not properly initialized."
            )

        n_samp = _sum_inplace(_np.atleast_1d(oks.shape[0]))

        n_par = grad.shape[0]

        if out is None:
            out = _np.zeros(n_par, dtype=_np.complex128)

        if self._has_complex_parameters:
            if self._use_iterative:
                op = self._linear_operator(oks, n_samp)

                if self._x0 is None:
                    self._x0 = _np.zeros(n_par, dtype=_np.complex128)

                out[:], info = self._sparse_solver(
                    op,
                    grad,
                    x0=self._x0,
                    tol=self.sparse_tol,
                    maxiter=self.sparse_maxiter,
                )
                if info < 0:
                    raise RuntimeError("SR sparse solver did not converge.")

                self._x0 = out
            else:
                self._S = _np.matmul(oks.conj().T, oks, self._S)
                self._S = _sum_inplace(self._S)
                self._S /= float(n_samp)

                self._apply_preconditioning(grad)

                if self._lsq_solver == "Cholesky":
                    c, low = _cho_factor(self._S, check_finite=False)
                    out[:] = _cho_solve((c, low), grad)

                else:
                    out[:], residuals, self._last_rank, s_vals = _lstsq(
                        self._S,
                        grad,
                        cond=self._svd_threshold,
                        lapack_driver=self._lapack_driver,
                    )

                self._revert_preconditioning(out)

        else:
            if self._use_iterative:
                op = self._linear_operator(oks, n_samp)

                if self._x0 is None:
                    self._x0 = _np.zeros(n_par)

                out[:].real, info = self._sparse_solver(
                    op,
                    grad.real,
                    x0=self._x0,
                    tol=self.sparse_tol,
                    maxiter=self.sparse_maxiter,
                )
                if info < 0:
                    raise RuntimeError("SR sparse solver did not converge.")
                self._x0 = out.real
            else:
                self._S = _np.matmul(oks.conj().T, oks, self._S)
                self._S /= float(n_samp)

                self._apply_preconditioning(grad)

                if self._lsq_solver == "Cholesky":
                    c, low = _cho_factor(self._S, check_finite=False)
                    out[:].real = _cho_solve((c, low), grad)
                else:
                    out[:].real, residuals, self._last_rank, s_vals = _lstsq(
                        self._S.real,
                        grad.real,
                        cond=self._svd_threshold,
                        lapack_driver=self._lapack_driver,
                    )

                self._revert_preconditioning(out.real)

            out.imag.fill(0.0)

        if _n_nodes > 1:
            self._comm.bcast(out, root=0)
            self._comm.barrier()

        return out

    def _apply_preconditioning(self, grad):
        if self._scale_invariant_pc:
            # Even if S is complex, its diagonal elements should be real since it
            # is Hermitian.
            self._diag_S = _np.sqrt(self._S.diagonal().real)

            cutoff = 1.0e-10

            index = self._diag_S <= cutoff
            self._diag_S[index] = 1.0
            self._S[index, :].fill(0.0)
            self._S[:, index].fill(0.0)
            _np.fill_diagonal(self._S, 1.0)

            self._S /= _np.vdot(self._diag_S, self._diag_S)

            grad /= self._diag_S

        # Apply diagonal shift
        self._S += self._diag_shift * _np.eye(self._S.shape[0])

    @property
    def last_rank(self):
        return self._last_rank

    @property
    def last_covariance_matrix(self):
        return self._S

    def _revert_preconditioning(self, out):
        if self._scale_invariant_pc:
            out /= self._diag_S
        return out

    @property
    def scale_invariant_regularization(self):
        r"""bool: Whether to use the scale-invariant regularization as described by
        Becca and Sorella (2017), pp. 143-144.
        https://doi.org/10.1017/9781316417041
        """
        return self._scale_invariant_pc

    @scale_invariant_regularization.setter
    def scale_invariant_regularization(self, activate):
        assert activate is True or activate is False

        if activate and self._use_iterative:
            raise NotImplementedError(
                """Scale-invariant regularization is
                   not implemented for iterative solvers at the moment."""
            )

        self._scale_invariant_pc = activate

    def __repr__(self):
        rep = "SR(solver="

        if self._use_iterative:
            rep += "iterative:CG"
        else:
            rep += self._lsq_solver + ", diag_shift=" + str(self._diag_shift)
            if self._svd_threshold is not None:
                rep += ", threshold=" << self._svd_threshold

        rep += ", has_complex_parameters=" + str(self._has_complex_parameters) + ")"
        return rep

    def _linear_operator(self, oks, n_samp):
        n_par = oks.shape[1]
        shift = self._diag_shift
        oks_conj = oks.conjugate()

        if self._has_complex_parameters:

            def matvec(v):
                v_tilde = self._v_tilde
                res = self._res_t

                v_tilde = _np.matmul(oks, v, v_tilde) / float(n_samp)
                res = _np.matmul(v_tilde, oks_conj, res)
                res = _sum_inplace(res) + shift * v
                return res

        else:

            def matvec(v):
                v_tilde = self._v_tilde
                res = self._res_t

                v_tilde = _np.matmul(oks, v, v_tilde) / float(n_samp)
                res = _np.matmul(v_tilde, oks_conj, res)
                res = _sum_inplace(res) + shift * v

                return res.real

        return LinearOperator((n_par, n_par), matvec=matvec, rmatvec=matvec)

    @property
    def has_complex_parameters(self):
        return self._has_complex_parameters

    @property
    def onthefly(self):
        return False



import numpy as _np
from netket.legacy import random as _random
from numba import jit, int64, float64
from numba.experimental import jitclass


@jitclass([("local_states", float64[:]), ("size", int64), ("n_states", int64)])
class _LocalKernel:
    def __init__(self, local_states, size):
        self.local_states = _np.sort(_np.asarray(local_states, dtype=_np.float64))
        self.size = size
        self.n_states = self.local_states.size

    def transition(self, state, state_1, log_prob_corr):

        for i in range(state.shape[0]):
            state_1[i] = state[i]

            si = _random.randint(0, self.size, size=())
            rs = _random.randint(0, self.n_states - 1, size=())

            state_1[i, si] = self.local_states[
                rs + (self.local_states[rs] >= state[i, si])
            ]

        log_prob_corr.fill(0.0)

    def random_state(self, state):

        for i in range(state.shape[0]):
            for si in range(state.shape[1]):
                rs = _random.randint(0, self.n_states, size=())
                state[i, si] = self.local_states[rs]



import abc
import numpy as _np


class AbstractOptimizer(abc.ABC):
    """Abstract class for NetKet optimizers"""

    def init(self, n_par, has_complex_parameters):
        r"""Initializes the optimizer.

        Args:
            n_par (int): Number of parameters to be optimized.
            has_complex_parameters (bool): Whether the target function has complex weights.

        """
        pass

    @abc.abstractmethod
    def update(self, grad, pars):
        r"""Update the parameters using gradient information.

        Args:
            grad (array): Gradient to be used for the update.
            pars (array): Parameters to be updated.

        Returns:
            array: Updated parameters
        """
        pass

    @abc.abstractmethod
    def reset(self):
        r"""Resets the internal state of the optimizer."""



import numpy as _np
from netket.legacy import random as _random
from numba import jit, int64, float64


class _ExchangeKernel:
    def __init__(self, hilbert, clusters):
        self.clusters = clusters
        self._hilbert = hilbert

    @staticmethod
    @jit(nopython=True)
    def _transition(state, state_1, log_prob_corr, clusters):

        clusters_size = clusters.shape[0]

        for k in range(state.shape[0]):
            state_1[k] = state[k]

            # pick a random cluster
            cl = _random.randint(0, clusters_size, size=())

            # sites to be exchanged
            si = clusters[cl][0]
            sj = clusters[cl][1]

            state_1[k, si], state_1[k, sj] = state[k, sj], state[k, si]

        log_prob_corr[:] = 0.0

    def transition(self, state, state_1, log_prob_corr):
        return self._transition(state, state_1, log_prob_corr, self.clusters)

    def random_state(self, state):

        for i in range(state.shape[0]):
            self._hilbert.random_state(out=state[i])



from functools import singledispatch
from . import numpy


@singledispatch
def Momentum(machine, learning_rate, beta=0.9, l2reg=0):
    r"""Momentum-based Optimizer.
        The momentum update incorporates an exponentially weighted moving average
        over previous gradients to speed up descent
        `Qian, N. (1999) <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.57.5612&rep=rep1&type=pdf>`_.
        The momentum vector :math:`\mathbf{m}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function
        :math:`G(\mathbf{p})`, the updates for the parameter :math:`p_k` and
        corresponding component of the momentum :math:`m_k` are

        .. math:: m^\prime_k &= \beta m_k + (1-\beta)G_k(\mathbf{p})\\
        p^\prime_k &= \eta m^\prime_k


        Args:
           learning_rate (float): The learning rate :math:`\eta`
           beta (float): Momentum exponential decay rate, should be in [0,1].
           l2reg (float): The amount of L2 regularization.

        Examples:
           Momentum optimizer.

           >>> from netket.optimizer import Momentum
           >>> op = Momentum(learning_rate=0.01)
    """
    return numpy.Momentum(learning_rate, beta, l2reg)



import netket.legacy as nk
from netket.legacy.optimizer import SR

import numpy as np

import pytest

pytestmark = pytest.mark.legacy


def test_sr_no_segfault():
    """
    Tests the resolution of bug #317.
    """
    # Hilbert space of spins on the graph
    hi = nk.hilbert.Spin(s=0.5) ** 20
    machine = nk.machine.RbmSpin(alpha=1, hilbert=hi)
    sr = SR(machine)
    assert sr.last_covariance_matrix is None


def test_svd_threshold():
    """
    Test SVD threshold option of BDCSVD
    """
    with pytest.raises(
        ValueError,
        match="The svd_threshold option is available only for non-sparse solvers.",
    ):
        # Hilbert space of spins on the graph
        hi = nk.hilbert.Spin(s=0.5) ** 20
        machine = nk.machine.RbmSpin(alpha=1, hilbert=hi)
        SR(machine, use_iterative=True, svd_threshold=1e-3)



import jax


class _JaxLocalKernel:
    def __init__(self, local_states, size):
        self.local_states = jax.numpy.sort(jax.numpy.array(local_states))
        self.size = size
        self.n_states = self.local_states.size

    def transition(self, key, state):

        keys = jax.random.split(key, 2)
        si = jax.random.randint(keys[0], shape=(1,), minval=0, maxval=self.size)
        rs = jax.random.randint(keys[1], shape=(1,), minval=0, maxval=self.n_states - 1)

        return jax.ops.index_update(
            state, si, self.local_states[rs + (self.local_states[rs] >= state[si])]
        )

    def random_state(self, key, state):
        keys = jax.random.split(key, 2)

        rs = jax.random.randint(
            keys[1], shape=(self.size,), minval=0, maxval=self.n_states
        )

        return keys[0], self.local_states[rs]



from netket import (
    exact,
    callbacks,
    graph,
    hilbert,
    # logging,
    operator,
    #    optim,
    #    random,
    #    sampler,
    #    stats,
    utils,
    #    variational,
    #    _exact_dynamics,
    #    _vmc,
    #    _steadystate,
)

from . import (
    # hilbert,
    machine,
    sampler,
    optimizer,
    random,
    stats,
    _vmc,
    _qsr,
    _steadystate,
    logging,
)

from ._vmc import Vmc
from ._qsr import Qsr
from ._steadystate import SteadyState

from .vmc_common import (
    tree_map as _tree_map,
    trees2_map as _trees2_map,
)



from .wrap import Wrap
import jax
import jax.experimental.optimizers as jaxopt
from .. import AdaDelta, AdaGrad, AdaMax, AmsGrad, Momentum, RmsProp, Sgd

from netket.legacy.machine import Jax as JaxMachine


@AdaDelta.register(JaxMachine)
def _JaxAdaDelta(machine, rho=0.95, epscut=1.0e-7, l2reg=0):
    raise NotImplementedError("Optimizer not implemented yet in Jax")


@AdaGrad.register(JaxMachine)
def _JaxAdaGrad(machine, learning_rate=0.001, epscut=1.0e-7):
    return Wrap(machine, jaxopt.adagrad(learning_rate))


@AdaMax.register(JaxMachine)
def _JaxAdaMax(machine, alpha=0.001, beta1=0.9, beta2=0.999, epscut=1.0e-7):
    return Wrap(machine, jaxopt.adamax(alpha, beta1, beta2, epscut))


@AmsGrad.register(JaxMachine)
def _JaxAdaDelta(machine, rho=0.95, epscut=1.0e-7, l2reg=0):
    raise NotImplementedError("Optimizer not implemented yet in Jax")


@Momentum.register(JaxMachine)
def _JaxMomentum(machine, learning_rate, beta=0.9, l2reg=0):
    return Wrap(machine, jaxopt.momentum(learning_rate, beta))


@RmsProp.register(JaxMachine)
def _JaxRmsProp(machine, learning_rate=0.001, beta=0.9, epscut=1.0e-7):
    return Wrap(machine, jaxopt.rmsprop(learning_rate, beta, epscut))


@Sgd.register(JaxMachine)
def _JaxSgd(machine, learning_rate, l2reg=0, decay_factor=1.0):
    return Wrap(machine, jaxopt.sgd(learning_rate))


# SR
from .. import SR as _SR
from .stochastic_reconfiguration import SR


@_SR.register(JaxMachine)
def _JaxSR(
    machine,
    lsq_solver=None,
    diag_shift=0.01,
    use_iterative=True,
    svd_threshold=None,
    sparse_tol=None,
    sparse_maxiter=None,
    onthefly=True,
):
    return SR(
        machine,
        lsq_solver,
        diag_shift,
        use_iterative,
        svd_threshold,
        sparse_tol,
        sparse_maxiter,
        onthefly,
    )



from numba import jit
import math

import numpy as _np
from netket.legacy import random as _random


class _HamiltonianKernel:
    def __init__(self, hamiltonian):
        self._hamiltonian = hamiltonian
        self._sections = _np.empty(1, dtype=_np.int32)
        self._hamconn = self._hamiltonian.get_conn_flattened
        self._n_conn = self._hamiltonian.n_conn
        self._hilbert = hamiltonian.hilbert

    def transition(self, state, state_1, log_prob_corr):

        sections = self._sections
        sections = _np.empty(state.shape[0], dtype=_np.int32)
        vprimes = self._hamconn(state, sections)[0]

        self._choose(vprimes, sections, state_1, log_prob_corr)

        self._n_conn(state_1, sections)

        log_prob_corr -= _np.log(sections)

    def random_state(self, state):

        for i in range(state.shape[0]):
            self._hilbert.random_state(out=state[i])

    @staticmethod
    @jit(nopython=True)
    def _choose(states, sections, out, w):
        low_range = 0
        for i, s in enumerate(sections):
            n_rand = _random.randint(low_range, s, size=())
            out[i] = states[n_rand]
            w[i] = math.log(s - low_range)
            low_range = s



from ..abstract_optimizer import AbstractOptimizer


class Sgd(AbstractOptimizer):
    def __init__(self, learning_rate, l2reg=0, decay_factor=1.0):

        self._learning_rate = learning_rate
        self._l2reg = l2reg
        self._decay_factor = decay_factor
        self._eta = learning_rate

        if learning_rate <= 0:
            raise ValueError("Invalid learning rate.")
        if l2reg < 0:
            raise ValueError("Invalid L2 regularization.")
        if decay_factor < 1:
            raise ValueError("Invalid decay factor.")

    def update(self, grad, pars):
        self._eta *= self._decay_factor
        pars -= (grad + self._l2reg * pars) * self._eta
        return pars

    def reset(self):
        self._eta = self._learning_rate

    def __repr__(self):
        if self._learning_rate != self._eta:
            lr = "learning_rate[base]={}, learning_rate[current]={}".format(
                self._learning_rate, self._eta
            )
        else:
            lr = "learning_rate={}".format(self._learning_rate)
        return ("Sgd({}, l2reg={}, decay_factor={})").format(
            lr, self._l2reg, self._decay_factor
        )



import math

from numba import jit
import numpy as _np
from . import mean as _mean
from . import var as _var
from . import total_size as _total_size


def _format_decimal(value, std, var):
    if math.isfinite(std) and std > 1e-7:
        decimals = max(int(_np.ceil(-_np.log10(std))), 0)
        return (
            "{0:.{1}f}".format(value, decimals + 1),
            "{0:.{1}f}".format(std, decimals + 1),
            "{0:.{1}f}".format(var, decimals + 1),
        )
    else:
        return (
            "{0:.3e}".format(value),
            "{0:.3e}".format(std),
            "{0:.3e}".format(var),
        )


class Stats:
    """A dict-compatible class containing the result of the statistics function."""

    _NaN = float("NaN")

    def __init__(
        self,
        mean=_NaN,
        error_of_mean=_NaN,
        variance=_NaN,
        tau_corr=_NaN,
        R_hat=_NaN,
    ):
        self.mean = complex(mean) if _np.iscomplexobj(mean) else float(mean)
        self.error_of_mean = float(error_of_mean)
        self.variance = float(variance)
        self.tau_corr = float(tau_corr)
        self.R_hat = float(R_hat)

    def to_json(self):
        jsd = {}
        jsd["Mean"] = self.mean.real
        jsd["Variance"] = self.variance
        jsd["Sigma"] = self.error_of_mean
        jsd["R_hat"] = self.R_hat
        jsd["TauCorr"] = self.tau_corr
        return jsd

    def __repr__(self):
        mean, err, var = _format_decimal(self.mean, self.error_of_mean, self.variance)
        if not math.isnan(self.R_hat):
            ext = ", R̂={:.4f}".format(self.R_hat)
        else:
            ext = ""
        return "{} ± {} [σ²={}{}]".format(mean, err, var, ext)

    def __getitem__(self, name):
        if name in ("mean", "Mean"):
            return self.mean
        elif name in ("variance", "Variance"):
            return self.variance
        elif name in ("error_of_mean", "Sigma"):
            return self.error_of_mean
        elif name in ("R_hat", "R"):
            return self.R_hat
        elif name in ("tau_corr", "TauCorr"):
            return self.tau_corr


@jit(nopython=True)
def _get_blocks(data, l):
    n_blocks = int(_np.floor(data.shape[1] / float(l)))
    blocks = _np.empty(data.shape[0] * n_blocks, dtype=data.dtype)
    k = 0
    for i in range(data.shape[0]):
        for b in range(n_blocks):
            blocks[k] = data[i, b * l : (b + 1) * l].mean()
            k += 1
    return blocks


def _block_variance(data, l):
    blocks = _get_blocks(data, l)
    ts = _total_size(blocks)
    if ts > 0:
        return _var(blocks), ts
    else:
        return _np.nan, 0


def _batch_variance(data):
    b_means = _np.mean(data, axis=1)
    ts = _total_size(b_means)
    return _var(b_means), ts


def statistics(data):
    r"""
    Returns statistics of a given array (or matrix, see below) containing a stream of data.
    This is particularly useful to analyze Markov Chain data, but it can be used
    also for other type of time series.

    Args:
        data (vector or matrix): The input data. It can be real or complex valued.
                                * if a vector, it is assumed that this is a time
                                  series of data (not necessarily independent).
                                * if a matrix, it is assumed that that rows data[i]
                                  contain independent time series.

    Returns:
       Stats: A dictionary-compatible class containing the average (mean),
             the variance (variance),
             the error of the mean (error_of_mean), and an estimate of the
             autocorrelation time (tau_corr). In addition to accessing the elements with the standard
             dict sintax (e.g. res['mean']), one can also access them directly with the dot operator
             (e.g. res.mean).
    """

    stats = Stats()
    data = _np.atleast_1d(data)
    if data.ndim == 1:
        data = data.reshape((1, -1))

    if data.ndim > 2:
        raise NotImplementedError("Statistics are implemented only for ndim<=2")

    mean = _mean(data)
    variance = _var(data)

    ts = _total_size(data)

    bare_var = variance

    batch_var, n_batches = _batch_variance(data)

    b_s = 32
    l_block = max(1, data.shape[1] // b_s)

    block_var, n_blocks = _block_variance(data, l_block)

    tau_batch = ((ts / n_batches) * batch_var / bare_var - 1) * 0.5
    tau_block = ((ts / n_blocks) * block_var / bare_var - 1) * 0.5

    block_good = n_blocks >= b_s and tau_block < 6 * l_block
    batch_good = n_batches >= b_s and tau_batch < 6 * data.shape[1]

    if batch_good:
        error_of_mean = _np.sqrt(batch_var / n_batches)
        tau_corr = max(0, tau_batch)
    elif block_good:
        error_of_mean = _np.sqrt(block_var / n_blocks)
        tau_corr = max(0, tau_block)
    else:
        error_of_mean = _np.nan
        tau_corr = _np.nan

    if n_batches > 1:
        N = data.shape[-1]

        # V_loc = _np.var(data, axis=-1, ddof=0)
        # W_loc = _np.mean(V_loc)
        # W = _mean(W_loc)
        # # This approximation seems to hold well enough for larger n_samples
        W = variance

        R_hat = _np.sqrt((N - 1) / N + batch_var / W)
    else:
        R_hat = float("nan")

    return Stats(mean, error_of_mean, variance, tau_corr, R_hat)



import numpy as np
import pytest

import netket.legacy as nk
import netket.vqs as vmc
from netket.operator import local_values
from netket.stats import statistics
from scipy.optimize import curve_fit
from numba import jit

from ... import common

pytestmark = pytest.mark.legacy


def _setup():
    g = nk.graph.Hypercube(3, 2)
    hi = nk.hilbert.Spin(0.5, N=g.n_nodes)

    ham = nk.operator.Heisenberg(hi, graph=g)

    ma = nk.machine.RbmSpin(hi, alpha=2)
    ma.init_random_parameters()

    return hi, ham, ma


def _test_stats_mean_std(hi, ham, ma, n_chains):
    sampler = nk.sampler.MetropolisLocal(ma, n_chains=n_chains)

    n_samples = 16000
    num_samples_per_chain = n_samples // n_chains

    # Discard a few samples
    sampler.generate_samples(1000)

    samples = sampler.generate_samples(num_samples_per_chain)
    assert samples.shape == (num_samples_per_chain, n_chains, hi.size)

    eloc = np.empty((num_samples_per_chain, n_chains), dtype=np.complex128)
    for i in range(num_samples_per_chain):
        eloc[i] = local_values(ham, ma, samples[i])

    stats = statistics(eloc.T)

    assert stats.mean == pytest.approx(np.mean(eloc))
    if n_chains > 1:

        # variance == average sample variance over chains
        assert stats.variance == pytest.approx(np.var(eloc))
        # R estimate
        B_over_n = stats.error_of_mean ** 2
        W = stats.variance
        assert stats.R_hat == pytest.approx(np.sqrt(1.0 + B_over_n / W), abs=1e-3)


@common.skipif_mpi
def test_stats_mean_std():
    hi, ham, ma = _setup()
    for bs in (1, 2, 16, 32):
        _test_stats_mean_std(hi, ham, ma, bs)


def _test_tau_corr(batch_size, sig_corr):
    def next_pow_two(n):
        i = 1
        while i < n:
            i = i << 1
        return i

    def autocorr_func_1d(x, norm=True):
        x = np.atleast_1d(x)
        if len(x.shape) != 1:
            raise ValueError("invalid dimensions for 1D autocorrelation function")
        n = next_pow_two(len(x))

        # Compute the FFT and then (from that) the auto-correlation function
        f = np.fft.fft(x - np.mean(x), n=2 * n)
        acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
        acf /= 4 * n

        # Optionally normalize
        if norm:
            acf /= acf[0]

        return acf

    @jit
    def gen_data(n_samples, log_f, dx, seed=1234):
        np.random.seed(seed)
        # Generates data with a simple markov chain
        x = np.empty(n_samples)
        x_old = np.random.normal()
        for i in range(n_samples):
            x_new = x_old + np.random.normal(scale=dx, loc=0.0)
            if np.exp(log_f(x_new) - log_f(x_old)) > np.random.uniform(0, 1):
                x[i] = x_new
            else:
                x[i] = x_old

            x_old = x[i]
        return x

    @jit
    def log_f(x):
        return -(x ** 2.0) / 2.0

    def func_corr(x, tau):
        return np.exp(-x / (tau))

    n_samples = 8000000 // batch_size

    data = np.empty((batch_size, n_samples))
    tau_fit = np.empty((batch_size))

    for i in range(batch_size):
        data[i] = gen_data(n_samples, log_f, sig_corr, seed=i + batch_size)
        autoc = autocorr_func_1d(data[i])
        popt, pcov = curve_fit(func_corr, np.arange(40), autoc[0:40])
        tau_fit[i] = popt[0]

    tau_fit_m = tau_fit.mean()

    stats = statistics(data)

    assert np.mean(data) == pytest.approx(stats.mean)
    assert np.var(data) == pytest.approx(stats.variance)

    assert tau_fit_m == pytest.approx(stats.tau_corr, rel=1, abs=3)

    eom_fit = np.sqrt(np.var(data) * tau_fit_m / float(n_samples * batch_size))

    print(stats.error_of_mean, eom_fit)
    assert eom_fit == pytest.approx(stats.error_of_mean, rel=0.6)


def test_tau_corr():
    sig_corr = 0.5
    for bs in (1, 2, 32, 64):
        _test_tau_corr(bs, sig_corr)


def test_decimal_format():
    from netket.stats import Stats

    assert str(Stats(1.0, 1e-3)) == "1.0000 ± 0.0010 [σ²=nan]"
    assert str(Stats(1.0, 1e-6)) == "1.0000000 ± 0.0000010 [σ²=nan]"
    assert str(Stats(1.0, 1e-7)) == "1.000e+00 ± 1.000e-07 [σ²=nan]"

    assert str(Stats(float("nan"), float("inf"))) == "nan ± inf [σ²=nan]"
    assert str(Stats(1.0, float("nan"))) == "1.000e+00 ± nan [σ²=nan]"
    assert str(Stats(1.0, float("inf"))) == "1.000e+00 ± inf [σ²=nan]"
    assert str(Stats(float("inf"), 0.0)) == "inf ± 0.000e+00 [σ²=nan]"
    assert str(Stats(1.0, 0.0)) == "1.000e+00 ± 0.000e+00 [σ²=nan]"

    assert str(Stats(1.0, 0.12, 0.5)) == "1.00 ± 0.12 [σ²=0.50]"
    assert str(Stats(1.0, 0.12, 0.5, R_hat=1.01)) == "1.00 ± 0.12 [σ²=0.50, R̂=1.0100]"



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class RmsProp(AbstractOptimizer):
    r"""RMSProp optimizer.

    RMSProp is a well-known update algorithm proposed by Geoff Hinton
    in his Neural Networks course notes `Neural Networks course notes
    <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.
    It corrects the problem with AdaGrad by using an exponentially weighted
    moving average over past squared gradients instead of a cumulative sum.
    After initializing the vector :math:`\mathbf{s}` to zero, :math:`s_k` and t
    he parameters :math:`p_k` are updated as

    .. math:: s^\prime_k = \beta s_k + (1-\beta) G_k(\mathbf{p})^2 \\
              p^\prime_k = p_k - \frac{\eta}{\sqrt{s_k}+\epsilon} G_k(\mathbf{p})

    """

    def __init__(self, learning_rate=0.001, beta=0.9, epscut=1.0e-7):
        r"""
        Constructs a new ``RmsProp`` optimizer.

        Args:
            learning_rate: The learning rate :math:`\eta`
            beta: Exponential decay rate.
            epscut: Small cutoff value.

        Examples:
            RmsProp optimizer.

            >>> from netket.optimizer import RmsProp
            >>> op = RmsProp(learning_rate=0.02)
        """

        if epscut <= 0:
            raise ValueError("Invalid epsilon cutoff.")
        if learning_rate < 0:
            raise ValueError("Invalid learning rate.")
        if beta < 0 or beta > 1:
            raise ValueError("Invalid beta.")

        self._eta = learning_rate
        self._beta = beta
        self._epscut = epscut

        self._mt = None

    def update(self, grad, pars):
        if self._mt is None:
            self._mt = _np.zeros(pars.shape[0])

        self._mt = self._beta * self._mt + (1.0 - self._beta) * _np.abs(grad) ** 2

        pars -= self._eta * grad / _np.sqrt(self._mt + self._epscut)

        return pars

    def reset(self):
        if self._mt is not None:
            self._mt.fill(0.0)

    def __repr__(self):
        rep = "AmsGrad optimizer with these parameters :"
        rep += "\nLearning rate = " + str(self._eta)
        rep += "\nbeta = " + str(self._beta)
        rep += "\nepscut = " + str(self._epscut)
        return rep



import numpy as _np
from ..abstract_sampler import AbstractSampler

from netket.legacy.stats import mean as _mean

from numba import jit
from numba import int64, float64
from netket.legacy import random as _random
import math


class MetropolisHastingsPt(AbstractSampler):
    def __init__(
        self,
        machine,
        kernel,
        n_replicas=32,
        sweep_size=None,
    ):
        super().__init__(machine)

        self.n_replicas = n_replicas

        self.sweep_size = sweep_size

        self._kernel = kernel

        self.machine_pow = 2.0

        # Contains quantities to compute diffusion coefficient of replicas
        # beta_stats[0] = position of replica at beta=1 (do not reset this!)
        # beta_stats[1] = running average of beta=1 position
        # beta_stats[2] = running average of beta=1 position**2
        # beta_stats[3] = current number of exchange steps performed
        self._beta_stats = _np.zeros(4)

        self.reset(True)

    @property
    def n_replicas(self):
        return self._n_replicas

    @n_replicas.setter
    def n_replicas(self, n_replicas):
        if n_replicas < 1:
            raise ValueError("Expected n_replicas>0. ")

        self._n_replicas = n_replicas

        self._state = _np.zeros((n_replicas, self._input_size))
        self._state1 = _np.copy(self._state)

        self._log_values = _np.zeros(n_replicas, dtype=_np.complex128)
        self._log_values_1 = _np.zeros(n_replicas, dtype=_np.complex128)
        self._log_prob_corr = _np.zeros(n_replicas)

        # Linearly spaced inverse temperature
        self._beta = _np.empty(n_replicas)

        for i in range(n_replicas):
            self._beta[i] = 1.0 - float(i) / float(n_replicas)

        # some temporary arrays
        self._proposed_beta = _np.empty(n_replicas)
        self._beta_prob = _np.empty(n_replicas)

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self._machine_pow = m_power

    @property
    def sweep_size(self):
        return self._sweep_size

    @sweep_size.setter
    def sweep_size(self, sweep_size):
        self._sweep_size = sweep_size if sweep_size != None else self._input_size
        if self._sweep_size < 0:
            raise ValueError("Expected a positive integer for sweep_size ")

    def reset(self, init_random=False):
        if init_random:
            self._kernel.random_state(self._state)

        self._log_values = self.machine.log_val(self._state, out=self._log_values)

        self._accepted_samples = _np.zeros(self._n_replicas)
        self._total_samples = 0

        self._beta_stats[1:].fill(0)

    @staticmethod
    @jit(nopython=True)
    def _fixed_beta_acceptance_kernel(
        state,
        state1,
        log_values,
        log_values_1,
        log_prob_corr,
        machine_pow,
        beta,
        accepted,
    ):

        for i in range(state.shape[0]):
            prob = math.exp(
                machine_pow
                * beta[i]
                * (log_values_1[i] - log_values[i] + log_prob_corr[i]).real
            )

            assert not math.isnan(prob)

            accept = prob > _random.uniform(0, 1)

            if accept:
                log_values[i] = log_values_1[i]
                state[i] = state1[i]
                accepted[i] += 1

    def __next__(self):

        _log_val = self.machine.log_val
        _fb_acc_kernel = self._fixed_beta_acceptance_kernel
        _state = self._state
        _state1 = self._state1
        _log_values = self._log_values
        _log_values_1 = self._log_values_1
        _log_prob_corr = self._log_prob_corr
        _machine_pow = self._machine_pow
        _accepted_samples = self._accepted_samples
        _t_kernel = self._kernel.transition
        _beta = self._beta
        _beta_stats = self._beta_stats
        _proposed_beta = self._proposed_beta
        _beta_prob = self._beta_prob

        for sweep in range(self.sweep_size):

            # Propose a new state using the transition kernel
            _t_kernel(_state, _state1, _log_prob_corr)

            _log_values_1 = _log_val(_state1, out=_log_values_1)

            # Acceptance Kernel for fixed-beta moves
            _fb_acc_kernel(
                _state,
                _state1,
                _log_values,
                _log_values_1,
                _log_prob_corr,
                _machine_pow,
                _beta,
                _accepted_samples,
            )

            # Transition + Acceptance Kernel for replica exchange moves
            self._exchange_step_kernel(
                _log_values,
                _machine_pow,
                _beta,
                _proposed_beta,
                _beta_prob,
                _beta_stats,
                _accepted_samples,
            )

        self._total_samples += self.sweep_size
        return self._state[_np.intp(_beta_stats[0])].reshape(1, -1)

    @staticmethod
    @jit(nopython=True)
    def _exchange_step_kernel(
        log_values, machine_pow, beta, proposed_beta, prob, beta_stats, accepted_samples
    ):
        # Choose a random swap order (odd/even swap)
        swap_order = _random.randint(0, 2, size=()).item()

        n_replicas = beta.shape[0]

        for i in range(swap_order, n_replicas, 2):
            inn = (i + 1) % n_replicas
            proposed_beta[i] = beta[inn]
            proposed_beta[inn] = beta[i]

        for i in range(n_replicas):
            prob[i] = math.exp(
                machine_pow * (proposed_beta[i] - beta[i]) * log_values[i].real
            )

        for i in range(swap_order, n_replicas, 2):
            inn = (i + 1) % n_replicas

            prob[i] *= prob[inn]

            if prob[i] > _random.uniform(0, 1):
                # swapping status
                beta[i], beta[inn] = beta[inn], beta[i]
                accepted_samples[i], accepted_samples[inn] = (
                    accepted_samples[inn],
                    accepted_samples[i],
                )

                if beta_stats[0] == i:
                    beta_stats[0] = inn
                elif beta_stats[0] == inn:
                    beta_stats[0] = i

        # Update statistics to compute diffusion coefficient of replicas
        # Total exchange steps performed
        beta_stats[-1] += 1

        delta = beta_stats[0] - beta_stats[1]
        beta_stats[1] += delta / float(beta_stats[-1])
        delta2 = beta_stats[0] - beta_stats[1]
        beta_stats[2] += delta * delta2

    @property
    def stats(self):
        stats = {}
        accept = self._accepted_samples / float(self._total_samples)

        stats["mean_acceptance"] = _mean(accept)
        stats["min_acceptance"] = _mean(accept.min())
        stats["max_acceptance"] = _mean(accept.max())

        # Average position of beta=1
        # This is normalized and centered around zero
        # In the ideal case the average should be zero
        stats["normalized_beta=1_position"] = (
            self._beta_stats[1] / float(self._n_replicas - 1) - 0.5
        )

        # Average variance on the position of beta=1
        # In the ideal case this quantity should be of order ~ [0.2, 1]
        stats["normalized_beta=1_diffusion"] = _np.sqrt(
            self._beta_stats[2] / self._beta_stats[-1]
        ) / float(self._n_replicas)

        return stats
